import express from 'express';
import { IdentityManager } from '..';
import { Logger } from '../log/logger';
import { MetricsStorage } from '../metrics/metricsStorage';
import { IAppOptions } from '../model/appOptions';
import { IConfiguration } from '../model/configuration';
import { IErrorScopeOptions } from '../model/errorScope';
export declare class BacktraceService<T extends {}> {
    private readonly name;
    private readonly port;
    private readonly _opts?;
    /**
     * Returns ArgumentParser instance arguments.
     */
    get arguments(): {
        configuration?: string | undefined;
    } | undefined;
    private static _serviceVersion;
    /**
     * Returns service version
     * @returns service version from the process.env variable
     */
    static get serviceVersion(): string | undefined;
    /**
     * Returns Backtrace-service version
     */
    static get backtraceServiceVersion(): string;
    /**
     * Create and startup express application with basic options.
     * By using this method, application will apply all
     * Backtrace service rules
     * @param name service name (required by service layer)
     * @param port service port (required by service layer). Note this
     * service requested port. Service layer might assign a different
     * service port based on the service descriptor.
     * @param opts optional application options
     * @param expressApp - deprecated. Optional express app instance.
     * Please use expressInstance from app instead.
     * By using this parameter you can create express app in different
     * framework and still be able to use Backtrace integrations
     */
    static bootstrap<T extends {}>(name: string, port: number, opts?: IAppOptions<T>, app?: express.Express): BacktraceService<T>;
    /**
     * Adds the healthcheck endpoints to the service. Healthchecks allow
     * to verify the service status and confirm the backtrace-node integration
     * setup.
     * @param expressApp current express app
     * @param apiPrefix api prefix
     * @param identityManager identity manager
     * @param options options
     */
    static setupHealthCheck(expressApp: express.Express, apiPrefix: string, identityManager: IdentityManager, opts: {
        healthcheck?: () => Promise<unknown>;
    }): void;
    /**
     * Read and parse service configuration
     * @param name service name
     * @param port service port number
     * @param configurationPath (optional) service configuration path
     * @param ignoreDescriptor (optional) determine if service
     * should create/read service layer service descriptor
     */
    static readConfiguration<T = {}>(name: string, port: number, configurationPath?: string, ignoreDescriptor?: boolean, packageName?: string): IConfiguration<T>;
    /**
     * Setup popular Backtrace middlewares
     * @param app express app
     */
    static setupMiddlewares(app: express.Express, prefix?: string, identityManager?: IdentityManager): void;
    static setupErrorScopeMiddleware(app: express.Express, opts?: IErrorScopeOptions): void;
    /**
     * Setup express error middleware
     * @param app express app
     */
    static setupErrorMiddleware(app: express.Express, identityManager?: IdentityManager, overrideLogger?: Logger): void;
    /**
     * Setup security middlewares
     * @param app express app
     */
    static setupAuth(app: express.Express): void;
    /**
     * Setup service layer integration
     * @param app express app
     */
    static setupServiceLayer(app: express.Express, name?: string, port?: number, coronerdCallback?: (param: string) => void): void;
    private static _service;
    readonly configuration: IConfiguration<T>;
    readonly coronerdAddress: Set<string>;
    /**
     * Api prefix. If your application is using service layer integration, apiPrefix will be
     * loaded from service descriptor, otherwise apiPrefix will be api/{serviceName}
     */
    readonly apiPrefix: string;
    identityManager: IdentityManager;
    readonly app: express.Express;
    readonly metrics: MetricsStorage;
    private _argumentParser;
    private _arguments?;
    private readonly _logger;
    private _descriptor;
    constructor(name: string, port: number, _opts?: IAppOptions<T> | undefined);
    /**
     * Returns the information if the service is available in the on-premise environment
     * @returns true if the environment is set to on-premise
     */
    isOnPremise(): boolean;
    start(clusterSupport?: boolean | number): void;
    /**
     * Initialize the service logger.
     * @returns Service logger instance.
     */
    private initializeLogger;
    /**
     * Initialize default arg parser object.
     */
    private initializeDefaultArgParser;
    /**
     * Determine if backtrace-service should use backtrace-node integration
     */
    private isBacktraceIntegrationEnabled;
    /**
     * Generate correct credentials for metricsStorage obejct based on configuration file.
     */
    private getMetricsSecret;
}
