"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ServiceConfiguration = void 0;
var fs_1 = __importDefault(require("fs"));
var path_1 = __importDefault(require("path"));
var __1 = require("..");
var packageUtils_1 = require("../utils/packageUtils");
/**
 * Backtrace service configuration. Class responsible for managing configuration paths/parsing them.
 */
var ServiceConfiguration = /** @class */ (function () {
    function ServiceConfiguration(_logger, _isExecutable) {
        if (_logger === void 0) { _logger = __1.Logger.getLogger(); }
        if (_isExecutable === void 0) { _isExecutable = !!process.pkg; }
        this._logger = _logger;
        this._isExecutable = _isExecutable;
        this._productionMode = 'production';
        this._runningInProdEnvironment = this._isExecutable || process.env.NODE_ENV === this._productionMode;
    }
    /**
     * @deprecated Please use instead getProjectSettingsPath. This method is deprecated because
     * of the way how we want to read configuration files. This method will return only one
     * path to service configuration, that might be wrong for complex environment when user want to
     * override default configuration.
     *
     * Select default path for service configuraiton path in different environment
     * By default Backtrace expect to run service from root service directory
     * by using npm commands. In environment different than production
     * @ServiceConfiguration class will read configuration file from root dir.
     * Otherwise from default Backtrace configuration directory
     */
    ServiceConfiguration.prototype.getProjectSettingsPath = function (serviceName) {
        if (this._runningInProdEnvironment) {
            return "/etc/backtrace/".concat(serviceName, "/").concat(serviceName, ".conf");
        }
        return path_1.default.join(process.cwd(), "".concat(serviceName, ".conf"));
    };
    /**
     * Select default paths for service configuraiton path in different environment
     * By default Backtrace expect to run service from root service directory
     * by using npm commands. In environment different than production
     * @ServiceConfiguration class will read configuration file from root dir.
     * Otherwise from default Backtrace configuration directory
     */
    ServiceConfiguration.prototype.getProjectSettingsPaths = function (serviceName, packageName) {
        if (!packageName) {
            packageName = this.getPackageName(serviceName);
        }
        var configurationPaths = [
            // path to a service package
            "/snapshot/".concat(packageName, "/").concat(serviceName, ".conf"),
            // path to a configuration path available in a not packaged app
            path_1.default.join(process.cwd(), "".concat(serviceName, ".conf")),
        ];
        if (this._runningInProdEnvironment) {
            configurationPaths.unshift("/etc/backtrace/".concat(serviceName, "/").concat(serviceName, ".conf"));
        }
        return configurationPaths;
    };
    /**
     * Loads service configuration file
     */
    ServiceConfiguration.prototype.loadConfiguration = function (configurationPaths) {
        if (!configurationPaths || configurationPaths.length === 0) {
            throw new Error("configurationPath is undefined");
        }
        for (var _i = 0, configurationPaths_1 = configurationPaths; _i < configurationPaths_1.length; _i++) {
            var configurationPath = configurationPaths_1[_i];
            // ignore undefined path or not existing configuration files.
            if (!configurationPath || !fs_1.default.existsSync(configurationPath)) {
                continue;
            }
            this._logger.info("Reading configuration file available ".concat(configurationPath));
            var config = JSON.parse(fs_1.default.readFileSync(configurationPath, 'utf8'));
            config.isProduction = process.env.NODE_ENV === 'production';
            if (config.ssl) {
                if (config.ssl.key) {
                    config.ssl.key = fs_1.default.readFileSync(config.ssl.key);
                }
                if (config.ssl.cert) {
                    config.ssl.cert = fs_1.default.readFileSync(config.ssl.cert);
                }
            }
            return config;
        }
        throw new Error("Cannot find configuration file." +
            "Backtrace-service configuration file paths: ".concat(configurationPaths));
    };
    /**
     * Returns a package name. If the package is running via npm command
     * and the process.env has a package name in the npm variables then
     * the package.json package name will be returned. Otherwise a service
     * name with the "backtrace-" prefix or name of the package available in
     * the package snapshot.
     */
    ServiceConfiguration.prototype.getPackageName = function (serviceName) {
        if (process.env.npm_package_name) {
            return process.env.npm_package_name;
        }
        return packageUtils_1.PackageUtils.isPackage
            ? packageUtils_1.PackageUtils.packageName
            : "backtrace-".concat(serviceName);
    };
    return ServiceConfiguration;
}());
exports.ServiceConfiguration = ServiceConfiguration;
//# sourceMappingURL=serviceConfiguration.js.map